using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Entity.Infrastructure;
using System.Data.Entity.ModelConfiguration.Configuration;
using System.Data.Objects;
using System.Linq;
using System.Reflection;
using System.Data.Entity;

namespace EntityContext.Core
{
    public abstract class ContextBase : DbContext
    {
        #region Fields

        private readonly DbContextAdapter adapter;

        #endregion Fields

        #region Ctors

        protected ContextBase(ContextSettings settings)
            : base(settings.ConnectionStringName)
        {
            this.Settings = settings;
            this.Entities = new List<EntityBase>();
            this.adapter = new DbContextAdapter(this);
        }

        public void Initalize(bool force = false)
        {
            // nb: consider thread safety!!
            if (force || ((this.IsInitalized = this.Database.Exists()) == false))
            {
                // force db creation
                this.adapter.SaveChanges();
            }

            this.IsInitalized = this.Database.Exists();
        }

        #endregion Ctors

        #region Properties

        public bool IsInitalized { get; private set; }

        public List<EntityBase> Entities { get; private set; }

        public ContextSettings Settings { get; private set; }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Creates a repository for type T.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public Repository<T> Repository<T>() where T : class
        {
            var result = this.adapter.Repository<T>();
            return result;
        }

        /// <summary>
        /// Creates the object set.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public IObjectSet<T> CreateObjectSet<T>() where T : class
        {
            var result = this.adapter.CreateObjectSet<T>();
            return result;
        }

        /// <summary>
        /// Changes the state of the object.
        /// </summary>
        /// <param name="entity">The entity.</param>
        /// <param name="entityState">State of the entity.</param>
        public void ChangeObjectState(object entity, EntityState entityState)
        {
            this.adapter.ChangeObjectState(entity, entityState);
        }

        /// <summary>
        /// Saves all changes made in this context to the underlying database.
        /// </summary>
        /// <returns>
        /// The number of objects written to the underlying database.
        /// </returns>
        /// <exception cref="T:System.InvalidOperationException">Thrown if the context has been disposed.</exception>
        public override int SaveChanges()
        {
            var result = this.adapter.SaveChanges();
            return result;
        }

        #endregion Methods

        #region Entity Registration

        protected override void OnModelCreating(DbModelBuilder modelBuilder)
        {
            var registrar = modelBuilder.Configurations;
            var entities = ContextBase.GetEntities(registrar);
            foreach (var entity in entities)
            {
                if (entity.BaseType == typeof(EntityBase))
                {
                    var instance = Activator.CreateInstance(entity) as EntityBase;
                    if (instance == null)
                    {
                        throw new EntityException(string.Format("Entity \"{0}\" should be descendent of \"{1}\""
                            , entity.FullName
                            , typeof(EntityBase).FullName));
                    }

                    this.Entities.Add(instance);
                }
            }

            this.Entities.ForEach(e => e.Register(modelBuilder));
        }

        /// <summary>
        /// Gets the entities.
        /// </summary>
        /// <param name="registrar">The registrar.</param>
        /// <returns></returns>
        /// <remarks>
        /// BEWARE: May need to be re-implemented when switching to another EF version (works with EntityFramework version 4.3.1.0).
        /// </remarks>
        private static IEnumerable<Type> GetEntities(ConfigurationRegistrar registrar)
        {
            var config = ContextBase.GetPrivateMemberValue(registrar, "_modelConfiguration");
            var entities = (List<Type>)ContextBase.GetPrivateMemberValue(config, "Entities", true);
            return entities;
        }

        private static object GetPrivateMemberValue(object instance, string name, bool isProperty = false)
        {
            var type = instance.GetType();
            var flag = BindingFlags.NonPublic | BindingFlags.Instance | (isProperty ? BindingFlags.GetProperty : BindingFlags.GetField);
            var memberInfo = type.GetMembers(flag);
            var result = default(object);
            if (isProperty)
            {
                var info = memberInfo.FirstOrDefault(m => m.Name == name) as PropertyInfo;
                result = info != null ? info.GetValue(instance, null) : null;
            }
            else
            {
                var info = memberInfo.FirstOrDefault(m => m.Name == name) as FieldInfo;
                result = info != null ? info.GetValue(instance) : null;
            }

            return result;
        }

        #endregion Entity Registration

        #region Helper(s)

        private class DbContextAdapter : IDisposable
        {
            #region Fields

            private readonly ContextBase context;

            #endregion Fields

            #region Ctors

            public DbContextAdapter(ContextBase context)
            {
                this.context = context;
            }

            #endregion Ctors

            #region Methods

            public int SaveChanges()
            {
                var objectContext = ((IObjectContextAdapter)this.context).ObjectContext;
                return objectContext.SaveChanges();
            }

            public IObjectSet<T> CreateObjectSet<T>() where T : class
            {
                var objectContext = ((IObjectContextAdapter)this.context).ObjectContext;
                var result = objectContext.CreateObjectSet<T>();
                return result;
            }

            public void ChangeObjectState(object entity, EntityState state)
            {
                var objectContext = ((IObjectContextAdapter)this.context).ObjectContext;
                objectContext.ObjectStateManager.ChangeObjectState(entity, state);
            }

            public Repository<T> Repository<T>() where T : class
            {
                var result = new Repository<T>(this.context);
                return result;
            }

            #endregion Methods

            #region IDisposable Implementation

            void IDisposable.Dispose()
            {
                var objectContext = ((IObjectContextAdapter)this.context).ObjectContext;
                objectContext.Dispose();
            }

            #endregion IDisposable Implementation
        }

        #endregion Helper(s)
    }
}