using System;
using System.Collections.Generic;
using System.Data.Entity;
using System.Data.Entity.Infrastructure;
using System.Data.Objects;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Transactions;

namespace EntityContext.Core
{
    public class ContextInitializer<T> : IDatabaseInitializer<T> where T : ContextBase
    {
        // model compatibility:
        // http://connect.microsoft.com/VisualStudio/feedback/details/731226/entity-framework-4-3-automatic-migrations-error-when-changing-class-name
        public void InitializeDatabase(T context)
        {
            if (context.IsInitalized == false)
            {
                using (new TransactionScope(TransactionScopeOption.Suppress))
                {
                    var objectContext = ((IObjectContextAdapter)context).ObjectContext;
                    if (string.IsNullOrEmpty(context.Settings.ScriptFileName) == false)
                    {
                        // create db script; when you're satisfied with the script, 
                        // set ScriptFileName to null or modify settings (app.config/web.config)
                        var script = objectContext.CreateDatabaseScript();
                        using (var writer = File.CreateText(context.Settings.ScriptFileName))
                        {
                            writer.Write(script);
                        }
                    }
                    else if (context.Database.Exists() == false)
                    {
                        // create and initialize database
                        context.Database.Create();

                        //Locator.Logger.Write("Database created successfully.", "Context");

                        this.FixCompatibilityIssues(context);
                        ContextInitializer<T>.AddConstraints(context);
                        ContextInitializer<T>.SeedEntityData(context);
                    }
                }
            }

            context.Entities.RemoveAll(e => true);
        }

        private void FixCompatibilityIssues(T context)
        {
            //if (context.Database.CompatibleWithModel(true))
            //{
            //}
            //var dbName = context.Database.Connection.Database;
            //var compatibilityCmd = string.Format("EXEC sp_dbcmptlevel '{0}', 80", dbName);
            //context.Database.ExecuteSqlCommand(compatibilityCmd);
        }

        private static void WriteSqlScript(string script, string filename)
        {
            var path = AppDomain.CurrentDomain.BaseDirectory;
            new List<string>
                {
                    "\\",
                    "Debug",
                    "Release",
                    "bin"
                }
                .ForEach(t =>
                    {
                        if (path.EndsWith(t))
                        {
                            var parent = Directory.GetParent(path);
                            path = parent.FullName;
                        }
                    });

            if (string.IsNullOrWhiteSpace(filename))
            {
                filename = Path.GetFileNameWithoutExtension(Process.GetCurrentProcess().MainModule.FileName);
            }

            if (string.IsNullOrWhiteSpace(filename))
            {
                throw new FileNotFoundException("Invalid filename.");
            }

            path = Path.Combine(path, filename);
            path = Path.ChangeExtension(path, "sql");
            using (var sw = File.CreateText(path))
            {
                sw.Write(script);
            }
        }

        private static void CreateTables(ObjectContext objectContext)
        {
            var dataBaseCreateScript = objectContext.CreateDatabaseScript();
            objectContext.ExecuteStoreCommand(dataBaseCreateScript);
        }

        private static void AddConstraints(T context)
        {
            if (context == null) { throw new ArgumentNullException("context"); }
            if (context.Entities == null) { throw new NullReferenceException("context.Entities"); }
            context.Entities.ForEach(e =>
            {
                var constraints = new List<string>();
                e.AddConstraint(constraints);
                try
                {
                    constraints.ForEach(constraint => context.Database.ExecuteSqlCommand(constraint));
                }
                catch (SqlException x)
                {
                    var errors = new StringBuilder();
                    for (var i = 0; i < x.Errors.Count; i++)
                    {
                        errors.AppendFormat("Source: {0}, Error: {1}, State: {2}, Class: {3}, Server: {4}, Message: {5}, Procedure: {6}, LineNumber: {7}",
                            x.Errors[i].Source, x.Errors[i].Number, x.Errors[i].State, x.Errors[i].Class, x.Errors[i].Server, x.Errors[i].Message, x.Errors[i].Procedure, x.Errors[i].LineNumber);
                        errors.AppendLine();
                    }

                    throw new Exception(errors.ToString());
                }
            });
        }

        private static void SeedEntityData(T context)
        {
            if (context == null) { throw new ArgumentNullException("context"); }
            if (context.Entities == null) { throw new NullReferenceException("context.Entities"); }
            context.Entities.ForEach(e =>
            {
                e.Seed(context);
                context.SaveChanges();
            });
        }
    }
}