using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Entity;
using System.Data.Objects;
using System.Linq;
using System.Linq.Expressions;

namespace EntityContext.Core
{
    public class Repository<T> where T : class
    {
        #region Fields

        private readonly IObjectSet<T> objectSet;
        private readonly ContextBase context;

        #endregion Fields

        #region Ctors

        public Repository(ContextBase context)
        {
            this.context = context;
            this.objectSet = this.context.CreateObjectSet<T>();
        }

        #endregion Ctors

        #region Methods

        public IQueryable<T> AsQueryable()
        {
            return this.objectSet;
        }

        public IEnumerable<T> GetAll(params Expression<Func<T, object>>[] includeProperties)
        {
            var query = Repository<T>.PerformInclusions(includeProperties, this.AsQueryable());
            return query.ToList();
        }

        public IEnumerable<T> Find(Expression<Func<T, bool>> where, params Expression<Func<T, object>>[] includeProperties)
        {
            var query = Repository<T>.PerformInclusions(includeProperties, this.AsQueryable());
            return query.Where(where);
        }

        public T Single(Expression<Func<T, bool>> where, params Expression<Func<T, object>>[] includeProperties)
        {
            var query = Repository<T>.PerformInclusions(includeProperties, this.AsQueryable());
            return query.Single(where);
        }

        public T First(Expression<Func<T, bool>> where, params Expression<Func<T, object>>[] includeProperties)
        {
            var query = Repository<T>.PerformInclusions(includeProperties, this.AsQueryable());
            return query.First(where);
        }

        public void Delete(T entity)
        {
            this.objectSet.DeleteObject(entity);
        }

        public void Insert(T entity)
        {
            this.objectSet.AddObject(entity);
        }

        public void Update(T entity)
        {
            this.objectSet.Attach(entity);
            this.context.ChangeObjectState(entity, EntityState.Modified);
        }

        #endregion Methods

        #region Helpers

        private static IQueryable<T> PerformInclusions(IEnumerable<Expression<Func<T, object>>> includeProperties
            , IQueryable<T> query)
        {
            return includeProperties.Aggregate(query, (current, includeProperty) => current.Include(includeProperty));
        }

        #endregion Helpers
    }
}